/*
     File: ElementsSortedByStateDataSource.m
 Abstract: Provides the table view data for the elements sorted by their standard physical state.
  Version: 1.11
 
 Disclaimer: IMPORTANT:  This Apple software is supplied to you by Apple
 Inc. ("Apple") in consideration of your agreement to the following
 terms, and your use, installation, modification or redistribution of
 this Apple software constitutes acceptance of these terms.  If you do
 not agree with these terms, please do not use, install, modify or
 redistribute this Apple software.
 
 In consideration of your agreement to abide by the following terms, and
 subject to these terms, Apple grants you a personal, non-exclusive
 license, under Apple's copyrights in this original Apple software (the
 "Apple Software"), to use, reproduce, modify and redistribute the Apple
 Software, with or without modifications, in source and/or binary forms;
 provided that if you redistribute the Apple Software in its entirety and
 without modifications, you must retain this notice and the following
 text and disclaimers in all such redistributions of the Apple Software.
 Neither the name, trademarks, service marks or logos of Apple Inc. may
 be used to endorse or promote products derived from the Apple Software
 without specific prior written permission from Apple.  Except as
 expressly stated in this notice, no other rights or licenses, express or
 implied, are granted by Apple herein, including but not limited to any
 patent rights that may be infringed by your derivative works or by other
 works in which the Apple Software may be incorporated.
 
 The Apple Software is provided by Apple on an "AS IS" basis.  APPLE
 MAKES NO WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION
 THE IMPLIED WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS
 FOR A PARTICULAR PURPOSE, REGARDING THE APPLE SOFTWARE OR ITS USE AND
 OPERATION ALONE OR IN COMBINATION WITH YOUR PRODUCTS.
 
 IN NO EVENT SHALL APPLE BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL
 OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) ARISING IN ANY WAY OUT OF THE USE, REPRODUCTION,
 MODIFICATION AND/OR DISTRIBUTION OF THE APPLE SOFTWARE, HOWEVER CAUSED
 AND WHETHER UNDER THEORY OF CONTRACT, TORT (INCLUDING NEGLIGENCE),
 STRICT LIABILITY OR OTHERWISE, EVEN IF APPLE HAS BEEN ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.
 
 Copyright (C) 2010 Apple Inc. All Rights Reserved.
 
 */

#import "ElementsSortedByStateDataSource.h"
#import "TheElementsAppDelegate.h"
#import "PeriodicElements.h"
#import "AtomicElement.h"
#import "AtomicElementTableViewCell.h"

@implementation ElementsSortedByStateDataSource

// ElementsDataSourceProtocol methods


// return the data used by the navigation controller and tab bar item

- (NSString *)name {
	return @"State";
}

- (NSString *)navigationBarName {
	return @"Grouped by State";
}

- (UIImage *)tabBarImage {
	return [UIImage imageNamed:@"state_gray.png"];
}

- (BOOL)showDisclosureIcon
{
	return YES;
}


// atomic state is displayed in a grouped style tableview
- (UITableViewStyle)tableViewStyle {
	return UITableViewStylePlain;
} 

// return the atomic element at the index 
- (AtomicElement *)atomicElementForIndexPath:(NSIndexPath *)indexPath {
	
	// this table has multiple sections. One for each physical state
	// [solid, liquid, gas, artificial]
	// the section represents the index in the state array
	// the row the index into the array of data for a particular state
	
	// get the state
	NSString *elementState = [[[PeriodicElements sharedPeriodicElements] elementPhysicalStatesArray] objectAtIndex:indexPath.section];
	
	// return the element in the state array
	return [[[PeriodicElements sharedPeriodicElements] elementsWithPhysicalState:elementState] objectAtIndex:indexPath.row];
}


// UITableViewDataSource methods

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath
{
	AtomicElementTableViewCell *cell = (AtomicElementTableViewCell *)[tableView dequeueReusableCellWithIdentifier:@"AtomicElementTableViewCell"];
	if (cell == nil) {
		cell = [[[AtomicElementTableViewCell alloc] initWithStyle:UITableViewCellStyleDefault reuseIdentifier:@"AtomicElementTableViewCell"] autorelease];
	}
    
	// configure cell contents
	// all the rows should show the disclosure indicator
	if ([self showDisclosureIcon])
		cell.accessoryType = UITableViewCellAccessoryDisclosureIndicator;
	
	// set the element for this cell as specified by the datasource. The atomicElementForIndexPath: is declared
	// as part of the ElementsDataSource Protocol and will return the appropriate element for the index row
	cell.element = [self atomicElementForIndexPath:indexPath];
	
	return cell;
}

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView  {
	// this table has multiple sections. One for each physical state
	// [solid, liquid, gas, artificial]
	// return the number of items in the states array
	return [[[PeriodicElements sharedPeriodicElements] elementPhysicalStatesArray] count];
}


- (NSInteger)tableView:(UITableView *)tableView  numberOfRowsInSection:(NSInteger)section {
	// this table has multiple sections. One for each physical state
	// [solid, liquid, gas, artificial]
	
	// get the state key for the requested section
	NSString *stateKey = [[[PeriodicElements sharedPeriodicElements] elementPhysicalStatesArray] objectAtIndex:section];
	
	// return the number of items that are in the array for that state
	return [[[PeriodicElements sharedPeriodicElements] elementsWithPhysicalState:stateKey] count];
}


- (NSString *)tableView:(UITableView *)tableView titleForHeaderInSection:(NSInteger)section {
	// this table has multiple sections. One for each physical state
	
	// [solid, liquid, gas, artificial]
	// return the state that represents the requested section
	// this is actually a delegate method, but we forward the request to the datasource in the view controller
	
	return [[[PeriodicElements sharedPeriodicElements] elementPhysicalStatesArray] objectAtIndex:section];
}


@end
